"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const ast_1 = require("../ast");
const errors_1 = require("../errors");
const helpers_1 = require("./helpers");
function registerOperators(state, ctx) {
    const { _$_vars, ops, makeScope, compileToken, } = ctx;
    ops.set(state.config.reservedNames.quote, function (r, args) {
        let compFnBody = '';
        _$_vars[ctx.varsCount] = r[1];
        compFnBody += `(_$_vars[${String(ctx.varsCount++)}])`;
        return compFnBody;
    });
    ops.set(state.config.reservedNames.self, function (r, args) {
        let compFnBody = '';
        compFnBody += `((_$_vars[0])(${args.map(x => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')}))`;
        return compFnBody;
    });
    ops.set('$__if', function (r, args) {
        // S expression: ($__if condition 't-expr 'f-expr)
        //  -> (if condition is true ) S expr  : t-expr
        //  -> (if condition is false) S expr  : f-expr
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__if', args, 2, 3);
        compFnBody += `(${compileToken(r, 1)}?(${compileToken([helpers_1.stripQuote(state, r[2])], 0)}):(${compileToken([helpers_1.stripQuote(state, r[3])], 0)}))`;
        return compFnBody;
    });
    ops.set('$__if-null', function (r, args) {
        // S expression: ($__if-null condition 'null-expr)
        //  -> (if condition is not null ) S expr  : condition
        //  -> (if condition is null)      S expr  : null-expr
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__if-null', args, 2, 2);
        compFnBody += `((()=>{let _$_rv=${compileToken(r, 1)};return _$_rv?_$_rv:(${compileToken([helpers_1.stripQuote(state, r[2])], 0)}});})())`;
        return compFnBody;
    });
    ops.set('$__cond', function (r, args) {
        // S expression: ($__cond 'cond1 'expr1 ... 'condN 'exprN)
        //  -> (if (eval condI) is true ) S expr  : exprI
        //  -> (if no matched)            S expr  : null
        let compFnBody = '';
        {
            errors_1.checkParamsLength('compileToken:$__cond', args, 1);
            compFnBody += `(`;
            for (let p = 0; p < args.length; p += 2) {
                compFnBody += `${compileToken([helpers_1.stripQuote(state, args[p])], 0)}?(${compileToken([helpers_1.stripQuote(state, args[p + 1])], 0)}):(`;
            }
            compFnBody += `null`;
            for (let p = 0; p < args.length; p += 2) {
                compFnBody += `)`;
            }
            compFnBody += `)`;
        }
        return compFnBody;
    });
    ops.set('$__while', function (r, args) {
        // S expression: ($__while 'condition 'expr1 ... 'exprN)
        //  -> (if condition is true at least 1 or more times) S expr  : exprN
        //  -> (else)                                          S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__while', args, 1);
        compFnBody += `((()=>{let _$_rv=null;while(${compileToken([helpers_1.stripQuote(state, r[1])], 0)}){_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}return _$_rv})())`;
        return compFnBody;
    });
    ops.set('$__do-while', function (r, args) {
        // S expression: ($__do-while 'condition 'expr1 ... 'exprN)
        //  -> (if condition is true at least 1 or more times) S expr  : exprN
        //  -> (else)                                          S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__do-until', args, 1);
        compFnBody += `((()=>{let _$_rv=null;do{_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}}while(${compileToken([helpers_1.stripQuote(state, r[1])], 0)})return _$_rv)())`;
        return compFnBody;
    });
    ops.set('$__until', function (r, args) {
        // S expression: ($__until 'condition 'expr1 ... 'exprN)
        //  -> (if condition is true at least 1 or more times) S expr  : exprN
        //  -> (else)                                          S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__until', args, 1);
        compFnBody += `((()=>{let _$_rv=null;while(!${compileToken([helpers_1.stripQuote(state, r[1])], 0)}){_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}return _$_rv})())`;
        return compFnBody;
    });
    ops.set('$__do-until', function (r, args) {
        // S expression: ($__do-until 'condition 'expr1 ... 'exprN)
        //  -> (if condition is true at least 1 or more times) S expr  : exprN
        //  -> (else)                                          S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__do-until', args, 1);
        compFnBody += `((()=>{let _$_rv=null;do{_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}}while(!${compileToken([helpers_1.stripQuote(state, r[1])], 0)})return _$_rv)())`;
        return compFnBody;
    });
    ops.set('$__repeat', function (r, args) {
        // S expression: ($__repeat 'i n-times 'expr1 ... 'exprN)
        //  -> (if n > 0) S expr  : exprN
        //  -> (else)     S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__repeat', args, 2);
        if (!ast_1.isSymbol(args[0])) {
            throw new Error(`[SX] compileToken: $__repeat : args[0] is not symbol.`);
        }
        makeScope(() => {
            const name = 'v' + ctx.varNamesCount++;
            ctx.varNames.set(args[0].symbol, name);
            compFnBody += `(((_$_n)=>{let _$_rv=null;for(let ${name}=0;${name}<_$_n;${name}++){_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}return _$_rv})(${compileToken(args, 1)}))`;
        });
        return compFnBody;
    });
    ops.set('$__for', function (r, args) {
        // S expression: ($__for 'x list 'expr1 ... 'exprN)
        //  -> (if list.length > 0) S expr  : exprN
        //  -> (else)               S expr  : null
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__for', args, 2);
        if (!ast_1.isSymbol(args[0])) {
            throw new Error(`[SX] compileToken: $__for : args[0] is not symbol.`);
        }
        makeScope(() => {
            const name = 'v' + ctx.varNamesCount++;
            ctx.varNames.set(args[0].symbol, name);
            compFnBody += `(((_$_l)=>{let _$_rv=null;for(const ${name} of _$_l){_$_rv=${r.slice(2).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}}return _$_rv})(${compileToken(args, 1)}))`;
        });
        return compFnBody;
    });
    ops.set('$__scope', function (r, args) {
        // S expression: ($__scope isBlockLocal returnMultiple '((name value) | name ...) 'expr1 ... 'exprN)
        //  -> (if returnMultiple)  S expr  : [expr1 ... exprN]
        //  -> (else)               S expr  : exprN
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__scope', args, 2);
        // r[1]: isBlockLocal
        // r[2]: returnMultiple
        if (!Array.isArray(r[3])) {
            throw new Error(`[SX] compileToken: $__scope : args[2] is not array.`);
        }
        makeScope(() => {
            for (const x of helpers_1.stripQuote(state, r[3])) {
                let name = '';
                if (Array.isArray(x)) {
                    if (x.length < 1) {
                        throw new Error(`[SX] compileToken: $__scope : args[0][?] is too short.`);
                    }
                    if (!ast_1.isSymbol(x[0])) {
                        throw new Error(`[SX] compileToken: $__scope : args[0][?][0] is not symbol.`);
                    }
                    name = x[0].symbol;
                    compFnBody += `(${'v' + ctx.varNamesCount}=${compileToken(x, 1)})`;
                }
                else {
                    if (!ast_1.isSymbol(x)) {
                        throw new Error(`[SX] compileToken: $__scope : args[0][?] is not symbol.`);
                    }
                    name = x.symbol;
                }
                ctx.varDefs += `var v${ctx.varNamesCount}=void 0;`;
                ctx.varNames.set(name, 'v' + ctx.varNamesCount++);
            }
            const s = `${r.slice(4).map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join(',')}`;
            compFnBody += r[2] ? `[${s}]` : `(${s})`;
        });
        return compFnBody;
    });
    ops.set('$__try', function (r, args) {
        // S expression: ($__try 'expr 'catch-expr)
        //  ->                               S expr  : expr
        //  -> (if error is raised in expr)  S expr  : catch-expr
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__try', args, 1, 2);
        compFnBody += `((()=>{try{${compileToken([helpers_1.stripQuote(state, r[1])], 0)}}catch(e${ctx.varNamesCount}){let e${ctx.varNamesCount + 1}=(_$_vars[${String(ctx.varsCount)}])();`;
        _$_vars[ctx.varsCount++] = helpers_1.getScope_stateApplied(state);
        makeScope(() => {
            ctx.varNames.set('$error', 'e' + ctx.varNamesCount++);
            ctx.varNames.set('$parent', 'e' + ctx.varNamesCount++);
            compFnBody += `${compileToken([helpers_1.stripQuote(state, r[2])], 0)}}})())`;
        });
        return compFnBody;
    });
    ops.set(state.config.reservedNames.raise, function (r, args) {
        // S expression: ($raise 'expr)
        //  -> S expr  : -
        let compFnBody = '';
        compFnBody += `((()=>{throw ${compileToken([helpers_1.stripQuoteOrPass(state, r[1])], 0)}})())`;
        return compFnBody;
    });
    ops.set('$boolean', function (r, args) {
        // S expression: ($boolean any)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$boolean', args, 1, 1);
        compFnBody += `((x0=${compileToken(args, 0)}),(Array.isArray(x0)&&x0.length===0?false:boolean(x0)))`;
        return compFnBody;
    });
    ops.set('$__get', function (r, args) {
        // S expression: ($__get 'nameOrIndex1 ... 'nameOrIndexN)
        //  -> S expr  : any
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__get', args, 1);
        const quoted = helpers_1.stripQuote(state, r[1]);
        const name = ast_1.isSymbol(quoted) ? quoted.symbol : (typeof quoted === 'string' ? quoted : null);
        if (typeof name !== 'string') {
            throw new Error(`[SX] compileToken: $__get : operand is not symbol: ${JSON.stringify(r[1])}.`);
        }
        let vName = '';
        if (ctx.varNames.has(name)) {
            vName = ctx.varNames.get(name);
        }
        else {
            _$_vars[ctx.varsCount] = helpers_1.resolveValueSymbol_dynamic(state, name);
            vName = `_$_vars[${String(ctx.varsCount++)}]`;
        }
        const regToVars = (symName) => {
            _$_vars[ctx.varsCount] = symName;
            return `_$_vars[${String(ctx.varsCount++)}]`;
        };
        compFnBody += `((${vName})${r.slice(2).map((x, idx, arr) => `[${ast_1.isSymbol(arr[idx]) ?
            regToVars(arr[idx].symbol) :
            compileToken(arr, idx)}]`).join('')})`;
        return compFnBody;
    });
    ops.set('$__let', function (r, args) {
        // S expression: ($__let 'nameStrOrSymbol expr)
        //  -> S expr  : any
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__let', args, 2, 2);
        const quoted = helpers_1.stripQuote(state, r[1]);
        const name = ast_1.isSymbol(quoted) ? quoted.symbol : (typeof quoted === 'string' ? quoted : null);
        if (typeof name !== 'string') {
            throw new Error(`[SX] compileToken: $__let : operand is not rvalue: ${JSON.stringify(r[1])}.`);
        }
        if (!ctx.varNames.has(name)) {
            ctx.varDefs += `var v${ctx.varNamesCount}=void 0;`;
            ctx.varNames.set(name, 'v' + ctx.varNamesCount++);
        }
        compFnBody += `(${ctx.varNames.get(name)}=${compileToken(r, 2)})`;
        return compFnBody;
    });
    ops.set('$__set', function (r, args) {
        // S expression: ($__set 'nameOrListOfNameOrIndex expr)
        //  -> S expr  : any
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__set', args, 2);
        const quoted = helpers_1.stripQuote(state, r[1]);
        const name = ast_1.isSymbol(quoted) ?
            quoted.symbol :
            (typeof quoted === 'string' ?
                quoted :
                (Array.isArray(quoted) ?
                    (ast_1.isSymbol(quoted[0]) ?
                        quoted[0].symbol :
                        (typeof quoted[0] === 'string' ? quoted[0] : null)) :
                    null));
        if (typeof name !== 'string') {
            throw new Error(`[SX] compileToken: $__set : operand is not rvalue: ${JSON.stringify(r[1])}.`);
        }
        let vName = '';
        if (ctx.varNames.has(name)) {
            vName = ctx.varNames.get(name);
        }
        else {
            _$_vars[ctx.varsCount] = helpers_1.resolveValueSymbol_dynamic(state, name);
            vName = `_$_vars[${String(ctx.varsCount++)}]`;
        }
        const regToVars = (symName) => {
            _$_vars[ctx.varsCount] = errors_1.checkUnsafeVarNames('compileToken:$__set', symName);
            return `_$_vars[${String(ctx.varsCount++)}]`;
        };
        let checkNameCalled = false;
        let checkNamePos = '';
        const checkName = (str) => {
            if (!checkNameCalled) {
                _$_vars[ctx.varsCount] = helpers_1.checkUnsafeVarNames_dynamic('compileToken:$__set');
                checkNamePos = String(ctx.varsCount++);
                checkNameCalled = true;
            }
            return `(_$_vars[${checkNamePos}](${str}))`;
        };
        compFnBody += `((${vName})${(Array.isArray(quoted) ? quoted.slice(1) : []).map((x, idx, arr) => `[${ast_1.isSymbol(arr[idx]) ?
            regToVars(arr[idx].symbol) :
            checkName(compileToken(arr, idx))}]`).join('')}=${compileToken(r, 2)})`;
        return compFnBody;
    });
    ops.set(state.config.reservedNames.not, function (r, args) {
        // S expression: ($not any)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$not', args, 1, 1);
        compFnBody += `(!${compileToken(args, 0)})`;
        return compFnBody;
    });
    ops.set('$__and', function (r, args) {
        // S expression: ($__and 'expr1 ... 'exprN)
        //  -> (if all of ($boolean expr1) ... ($boolean exprN) are true) S expr  : exprN
        //  -> (else)                                                     S expr  : expr-i (false left most)
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__and', args, 1);
        compFnBody += `(${args.map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join('&&')})`;
        return compFnBody;
    });
    ops.set('$__or', function (r, args) {
        // S expression: ($__or 'expr1 ... 'exprN)
        //  -> (if any ($boolean expr1) ... ($boolean exprN) are true) S expr  : expr-i (where i: index of item first ($boolean expr-i) is to be true)
        //  -> (else)                                                  S expr  : expr-i (false right most)
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$__or', args, 1);
        compFnBody += `(${args.map((x) => compileToken([helpers_1.stripQuote(state, x)], 0)).join('||')})`;
        return compFnBody;
    });
    ops.set('===', function (r, args) {
        // S expression: ($eq arg1 arg2)
        //  -> (if arg1 === arg2)  S expr  : true
        //  -> (else)              S expr  : false
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:===', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('===')})`;
        return compFnBody;
    });
    ops.set('!==', function (r, args) {
        // S expression: ($not-eq arg1 arg2)
        //  -> (if arg1 !== arg2)  S expr  : true
        //  -> (else)              S expr  : false
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:!==', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('!==')})`;
        return compFnBody;
    });
    ops.set('==', function (r, args) {
        // S expression: (== a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:==', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('==')})`;
        return compFnBody;
    });
    ops.set('!=', function (r, args) {
        // S expression: (!= a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:!=', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('!=')})`;
        return compFnBody;
    });
    ops.set('<', function (r, args) {
        // S expression: (< a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:<', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('<')})`;
        return compFnBody;
    });
    ops.set('<=', function (r, args) {
        // S expression: (<= a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:<=', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('<=')})`;
        return compFnBody;
    });
    ops.set('>', function (r, args) {
        // S expression: (> a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:>', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('>')})`;
        return compFnBody;
    });
    ops.set('>=', function (r, args) {
        // S expression: (>= a b)
        //  -> S expr  : boolean
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:<=', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('>=')})`;
        return compFnBody;
    });
    ops.set('$concat', function (r, args) {
        // S expression: ($concat listOrString1 ... listOrStringN)
        //  -> S expr  : listOrString
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$concat', args, 1);
        let w1 = '';
        if (Array.isArray(args[0]) && ast_1.isSymbol(args[0][0], state.config.reservedNames.spread)) {
            const w0 = compileToken(args[0], 1);
            w1 = `(${w0}[0]).concat((${w0}.length>1?${w0}[1]:(typeof ${w0}[0]==='string'?'':[])),`;
        }
        else {
            w1 = `${compileToken(args, 0)}.concat(`;
        }
        compFnBody += `(${w1}${args.slice(1).map((x, idx, arr) => compileToken(arr, idx)).join(',')}))`;
        return compFnBody;
    });
    ops.set('+', function (r, args) {
        // S expression: (+ number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:+', args, 1);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return _$_rv.reduce((x,y)=>x+y);})())`;
        }
        else {
            compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('+')})`;
        }
        return compFnBody;
    });
    ops.set('-', function (r, args) {
        // S expression: (- number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:-', args, 1);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return (_$_rv.length>1?(_$_rv.reduce((x,y)=>x-y)):(_$_rv.length>0?-_$_rv[0]:NaN));})())`;
        }
        else {
            compFnBody += `(${r.length > 2 ?
                args.map((x, idx, arr) => compileToken(arr, idx)).join('-') :
                `-(${String(compileToken(r, 1))})`})`;
        }
        return compFnBody;
    });
    ops.set('*', function (r, args) {
        // S expression: (* number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:*', args, 2);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return _$_rv.reduce((x,y)=>x*y);})())`;
        }
        else {
            compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('*')})`;
        }
        return compFnBody;
    });
    ops.set('**', function (r, args) {
        // S expression: (** number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:**', args, 2);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return _$_rv.reduce((x,y)=>x**y);})())`;
        }
        else {
            compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('**')})`;
        }
        return compFnBody;
    });
    ops.set('/', function (r, args) {
        // S expression: (/ number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:/', args, 2);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return _$_rv.reduce((x,y)=>x/y);})())`;
        }
        else {
            compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('/')})`;
        }
        return compFnBody;
    });
    ops.set('%', function (r, args) {
        // S expression: (% number1 ... numberN)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:%', args, 2);
        let hasSpread = false;
        args.map((x, idx, arr) => {
            if (Array.isArray(x) && ast_1.isSymbol(x[0], state.config.reservedNames.spread)) {
                hasSpread = true;
            }
        });
        if (hasSpread) {
            compFnBody += `((()=>{let _$_rv=[];${args.map((x, idx, arr) => `_$_rv.push(${compileToken(arr, idx)})`).join(';')};return _$_rv.reduce((x,y)=>x%y);})())`;
        }
        else {
            compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('%')})`;
        }
        return compFnBody;
    });
    ops.set('<<', function (r, args) {
        // S expression: (<< number shift)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:<<', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('<<')})`;
        return compFnBody;
    });
    ops.set('>>', function (r, args) {
        // S expression: (>> number shift)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:>>', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('>>')})`;
        return compFnBody;
    });
    ops.set('>>>', function (r, args) {
        // S expression: (>>> number shift)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:>>>', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('>>>')})`;
        return compFnBody;
    });
    ops.set('$bit-not', function (r, args) {
        // S expression: ($bit-not number)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$bit-not', args, 1, 1);
        compFnBody += `(~(${compileToken(r, 1)}))`;
        return compFnBody;
    });
    ops.set('$bit-and', function (r, args) {
        // S expression: ($bit-and numberA numberB)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$bit-and', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('&')})`;
        return compFnBody;
    });
    ops.set('$bit-or', function (r, args) {
        // S expression: ($bit-or numberA numberB)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$bit-or', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('|')})`;
        return compFnBody;
    });
    ops.set('$bit-xor', function (r, args) {
        // S expression: ($bit-xor numberA numberB)
        //  -> S expr  : number
        let compFnBody = '';
        errors_1.checkParamsLength('compileToken:$bit-xor', args, 2, 2);
        compFnBody += `(${args.map((x, idx, arr) => compileToken(arr, idx)).join('^')})`;
        return compFnBody;
    });
}
exports.registerOperators = registerOperators;
//# sourceMappingURL=index.js.map