"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../../types");
const consts_1 = require("../../consts");
const ast_1 = require("../../ast");
const evaluate_1 = require("../../evaluate");
const compile_1 = require("../../compile");
const errors_1 = require("../../errors");
exports.$car = (state, name) => (...args) => {
    // S expression: ($car '(first second ... last))
    //  -> S expr  : first
    errors_1.checkParamsLength('$car', args, 1, 1);
    const car = exports.$$first(...args);
    if (!Array.isArray(car)) {
        throw new Error(`[SX] $car: Invalid argument(s): args[0] is not array.`);
    }
    if (car.length === 0) {
        throw new Error(`[SX] $car: Invalid argument(s): args[0] is nil.`);
    }
    return car[0];
};
exports.$$car = exports.$car(null, null);
exports.$cdr = (state, name) => (...args) => {
    // S expression: ($cdr '(first second ... last))
    //  -> S expr  : (second ... last)
    errors_1.checkParamsLength('$cdr', args, 1, 1);
    const car = exports.$$first(...args);
    if (!Array.isArray(car)) {
        throw new Error(`[SX] $cdr: Invalid argument(s): args[0] is not array.`);
    }
    if (car.length === 0) {
        throw new Error(`[SX] $cdr: Invalid argument(s): args[0] is nil.`);
    }
    return car.slice(1);
};
exports.$$cdr = exports.$cdr(null, null);
exports.$cons = (state, name) => (...args) => {
    // S expression: ($cons arg1 '(arg2-item1 ...) ... argN)
    //  -> S expr  : (arg1 arg2-item1 ...)
    // S expression: ($cons arg1 nilOrNull ... argN)
    //  -> S expr  : (arg1)
    // S expression: ($cons arg1 arg2 ... argN)
    //  -> S expr  : arg1.arg2
    errors_1.checkParamsLength('$cons', args, 2, 2);
    let { car, cdr } = exports.$$firstAndSecond(...args);
    if (car === null) {
        car = [];
    }
    if (cdr === null) {
        cdr = [];
    }
    if (Array.isArray(cdr)) {
        cdr = cdr.slice(0);
        cdr.unshift(car);
        return cdr;
    }
    else {
        return { car, cdr };
    }
};
exports.$$cons = exports.$cons(null, null);
exports.$first = (state, name) => (...args) => {
    // S expression: ($first first second ... last)
    //  -> S expr  : first
    // S expression: ($first)
    //  -> S expr  : null
    const car = args.slice(0, 1);
    return (car.length === 1) ? car[0] : null;
};
exports.$$first = exports.$first(null, null);
exports.$second = (state, name) => (...args) => {
    // S expression: ($second first second ... last)
    //  -> S expr  : second
    // S expression: ($second first)
    //  -> S expr  : null
    const cdr = args.slice(1, 2);
    return (cdr.length === 1) ? cdr[0] : null;
};
exports.$$second = exports.$second(null, null);
exports.$last = (state, name) => (...args) => {
    // S expression: ($last first second ... last)
    //  -> S expr  : last
    // S expression: ($last)
    //  -> S expr  : null
    const car = args.slice(args.length - 1, args.length);
    return (car.length === 1) ? car[0] : null;
};
exports.$$last = exports.$last(null, null);
exports.$rest = (state, name) => (...args) => {
    // S expression: ($rest first second ... last)
    //  -> S expr  : (second ... last)
    // S expression: ($rest first)
    //  -> S expr  : null
    const cdr = args.slice(1);
    return (0 < cdr.length) ? cdr : null;
};
exports.$$rest = exports.$rest(null, null);
exports.$firstAndSecond = (state, name) => (...args) => {
    // S expression: ($first-and-second first second ... last)
    //  -> S expr  : first.second
    let car = args.slice(0, 1);
    car = (car.length === 1) ? car[0] : null;
    let cdr = args.slice(1, 2);
    cdr = (cdr.length === 1) ? cdr[0] : null;
    return { car, cdr };
};
exports.$$firstAndSecond = exports.$firstAndSecond(null, null);
exports.$atom = (state, name) => (...args) => {
    // S expression: ($atom arg1 ...)
    //  -> (if arg1 is list or dotted pair)  S expr  : false
    //  -> (if arg1 is nil or anything else) S expr  : true
    errors_1.checkParamsLength('$atom', args, 1, 1);
    const car = exports.$$first(...args);
    if (car === null || car === void 0) {
        return true;
    }
    if (Array.isArray(car)) {
        if (car.length === 0)
            return true;
        else
            return false;
    }
    switch (typeof car) {
        case 'number':
        case 'string':
        case 'function':
        case 'boolean':
            return true;
        case 'object':
            return ast_1.isSymbol(car) ? true : false;
    }
    return false;
};
exports.$$atom = exports.$atom(null, null);
exports.$eq = (state, name) => (...args) => {
    // S expression: ($eq arg1 arg2)
    //  -> (if arg1 === arg2)  S expr  : true
    //  -> (else)              S expr  : false
    errors_1.checkParamsLength('$eq', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return car === cdr;
};
exports.$$eq = exports.$eq(null, null);
exports.$notEq = (state, name) => (...args) => {
    // S expression: ($not-eq arg1 arg2)
    //  -> (if arg1 !== arg2)  S expr  : true
    //  -> (else)              S expr  : false
    errors_1.checkParamsLength('$notEq', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return car !== cdr;
};
exports.$$notEq = exports.$notEq(null, null);
exports.$list = (state, name) => (...args) => 
// S expression: ($list arg1 ... argN)
//  -> S expr  : (arg1 ... argN)
args.slice(0);
exports.$$list = exports.$list(null, null);
// tslint:disable-next-line:variable-name
exports.$__scope = (state, name, capturedScopes) => (...args) => {
    // S expression: ($__scope isBlockLocal returnMultiple '((name value) | name ...) 'expr1 ... 'exprN)
    //  -> (if returnMultiple)  S expr  : [expr1 ... exprN]
    //  -> (else)               S expr  : exprN
    errors_1.checkParamsLength('$__scope', args, 3);
    const isBlockLocal = exports.$$first(...args);
    const returnMultiple = exports.$$second(...args);
    const { car, cdr } = exports.$$firstAndSecond(...args.slice(2));
    let r = null;
    const scope = {};
    if (Array.isArray(car)) {
        for (const x of car) {
            if (Array.isArray(x)) {
                const kv = exports.$$firstAndSecond(...x);
                const kvSym = ast_1.isSymbol(kv.car);
                const kvName = kvSym ? kvSym.symbol : String(kv.car);
                errors_1.checkUnsafeVarNamesEx('$__scope', scope, kvName);
                scope[kvName] = evaluate_1.evaluate(state, kv.cdr);
            }
            else {
                const xSym = ast_1.isSymbol(x);
                const xName = xSym ? xSym.symbol : String(x);
                errors_1.checkUnsafeVarNamesEx('$__scope', scope, xName);
                scope[xName] = null;
            }
        }
    }
    evaluate_1.installScope(state, scope, isBlockLocal, capturedScopes);
    try {
        if (4 < args.length) {
            if (returnMultiple) {
                r = [];
                for (const x of args.slice(3)) {
                    r.push(evaluate_1.evaluate(state, x));
                }
            }
            else {
                for (const x of args.slice(3)) {
                    r = evaluate_1.evaluate(state, x);
                }
            }
        }
        else {
            r = evaluate_1.evaluate(state, cdr);
        }
    }
    finally {
        evaluate_1.uninstallScope(state);
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__globalScope = (state, name) => (...args) => {
    // S expression: ($__global returnMultiple 'expr1 ... 'exprN)
    //  -> (if returnMultiple)  S expr  : [expr1 ... exprN]
    //  -> (else)               S expr  : exprN
    errors_1.checkParamsLength('$__globalScope', args, 1);
    const returnMultiple = exports.$$first(...args);
    const cdr = exports.$$second(...args);
    let r = null;
    evaluate_1.installScope(state, evaluate_1.getGlobalScope(state).scope, true);
    try {
        if (2 < args.length) {
            if (returnMultiple) {
                r = [];
                for (const x of args.slice(1)) {
                    r.push(evaluate_1.evaluate(state, x));
                }
            }
            else {
                for (const x of args.slice(1)) {
                    r = evaluate_1.evaluate(state, x);
                }
            }
        }
        else {
            r = evaluate_1.evaluate(state, cdr);
        }
    }
    finally {
        evaluate_1.uninstallScope(state);
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__capture = (state, name) => (...args) => {
    // S expression: ($__capture '(sym1 ... symN) 'expr1 ... 'exprN)
    //  -> S expr  : exprN
    errors_1.checkParamsLength('$__capture', args, 1);
    const formalArgs = args[0];
    if (!Array.isArray(formalArgs)) {
        throw new Error(`[SX] $__lambda: Invalid argument(s): args[0] is not array.`);
    }
    let r = null;
    const capturedScopes = evaluate_1.collectCapturedVariables(state, formalArgs);
    evaluate_1.installScope(state, {}, true, capturedScopes);
    try {
        for (const x of args.slice(1)) {
            r = evaluate_1.evaluate(state, x);
        }
    }
    finally {
        evaluate_1.uninstallScope(state);
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__lambda = (state, name) => (...args) => {
    // S expression: ($__lambda '(sym1 ... symN) 'expr1 ... 'exprN)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$__lambda', args, 2);
    const formalArgs = args[0];
    if (!Array.isArray(formalArgs)) {
        throw new Error(`[SX] $__lambda: Invalid argument(s): args[0] is not array.`);
    }
    let lastIsSpread = false;
    for (let i = 0; i < formalArgs.length; i++) {
        const fa = formalArgs[i];
        if (i === formalArgs.length - 1 && state.config.enableSpread &&
            Array.isArray(fa) && ast_1.isSymbol(fa[0], state.config.reservedNames.spread)) {
            if (!ast_1.isSymbol(fa[1])) {
                throw new Error(`[SX] $__lambda: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
            }
            formalArgs[i] = fa[1];
            lastIsSpread = true;
        }
        else if (!ast_1.isSymbol(fa)) {
            throw new Error(`[SX] $__lambda: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
        }
    }
    let fnBody = args.slice(1);
    if (state.config.enableTailCallOptimization) {
        fnBody = evaluate_1.optimizeTailCall(state, formalArgs, fnBody);
    }
    const capturedScopes = evaluate_1.getCapturedScopes(state);
    const fn = function (...actualArgs) {
        if ((actualArgs.length + (lastIsSpread ? 1 : 0)) < formalArgs.length) {
            throw new Error(`[SX] func call: Actual args too short: actual ${actualArgs.length} / formal ${formalArgs.length}.`);
        }
        // TODO: add type checking
        return exports.$__scope(state, name, capturedScopes)(false, false, [
            [state.config.reservedNames.self, fn],
            [state.config.reservedNames.thiz, this === consts_1.globalObj || this === void 0 ? null : ast_1.quote(state, this)],
            ...(formalArgs.map((x, index) => [
                x.symbol,
                ast_1.quote(state, (lastIsSpread && index === formalArgs.length - 1) ?
                    actualArgs.slice(index) : actualArgs[index])
            ])),
        ], ...fnBody);
    };
    return fn;
};
// tslint:disable-next-line:variable-name
exports.$comp$__lambda = (state, name) => (...args) => {
    // S expression: ($__lambda '(sym1 ... symN) 'expr1 ... 'exprN)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$$__lambda', args, 2);
    const formalArgs = args[0];
    if (!Array.isArray(formalArgs)) {
        throw new Error(`[SX] $$__lambda: Invalid argument(s): args[0] is not array.`);
    }
    let lastIsSpread = false;
    for (let i = 0; i < formalArgs.length; i++) {
        const fa = formalArgs[i];
        if (i === formalArgs.length - 1 && state.config.enableSpread &&
            Array.isArray(fa) && ast_1.isSymbol(fa[0], state.config.reservedNames.spread)) {
            if (!ast_1.isSymbol(fa[1])) {
                throw new Error(`[SX] $$__lambda: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
            }
            formalArgs[i] = fa[1];
            lastIsSpread = true;
        }
        else if (!ast_1.isSymbol(fa)) {
            throw new Error(`[SX] $$__lambda: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
        }
    }
    const fnBody = args.slice(1);
    return compile_1.compileLambda(state, formalArgs, lastIsSpread, fnBody);
};
// tslint:disable-next-line:variable-name
exports.$__defun = (state, name) => (...args) => {
    // S expression: ($__defun 'name '(sym1 ... symN) 'expr ... 'expr)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$__defun', args, 3);
    const car = exports.$$first(...args);
    const fn = exports.$__lambda(state, name)(...args.slice(1));
    errors_1.checkUnsafeVarNamesEx('$__defun', state.funcMap, car.symbol);
    // TODO: overloading
    state.funcMap.set(car.symbol, {
        name: car.symbol,
        fn: (st, nm) => fn,
    });
    return fn;
};
// tslint:disable-next-line:variable-name
exports.$comp$__defun = (state, name) => (...args) => {
    // S expression: ($__defun 'name '(sym1 ... symN) 'expr ... 'expr)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$$__defun', args, 3);
    const car = exports.$$first(...args);
    const fn = exports.$comp$__lambda(state, name)(...args.slice(1));
    errors_1.checkUnsafeVarNamesEx('$$__defun', state.funcMap, car.symbol);
    // TODO: overloading
    state.funcMap.set(car.symbol, {
        name: car.symbol,
        fn: (st, nm) => fn,
    });
    return fn;
};
// tslint:disable-next-line:variable-name
exports.$__refun = (state, name) => (...args) => {
    // S expression: ($refun 'name)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$__refun', args, 1, 1);
    const car = exports.$$first(...args);
    errors_1.checkUnsafeVarNamesEx('$__refun', state.funcMap, car.symbol);
    const info = state.funcMap.get(car.symbol);
    if (!info) {
        throw new Error(`[SX] $__refun: function ${car.symbol} is not defined.`);
    }
    return info.fn(state, car.symbol);
};
// tslint:disable-next-line:variable-name
exports.$__defmacro = (state, name) => (...args) => {
    // S expression: ($__defmacro 'name '(sym1 ... symN) 'expr ... 'expr)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$__defmacro', args, 3);
    const car = exports.$$first(...args);
    const formalArgs = args[1];
    if (!Array.isArray(formalArgs)) {
        throw new Error(`[SX] $__defmacro: Invalid argument(s): args[1] is not array.`);
    }
    let lastIsSpread = false;
    for (let i = 0; i < formalArgs.length; i++) {
        const fa = formalArgs[i];
        if (i === formalArgs.length - 1 && state.config.enableSpread &&
            Array.isArray(fa) && ast_1.isSymbol(fa[0], state.config.reservedNames.spread)) {
            if (!ast_1.isSymbol(fa[1])) {
                throw new Error(`[SX] $__defmacro: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
            }
            formalArgs[i] = fa[1];
            lastIsSpread = true;
        }
        else if (!ast_1.isSymbol(fa)) {
            throw new Error(`[SX] $__defmacro: Invalid formal argument(s): item(s) of args[${i}] is not symbol.`);
        }
    }
    const fnBody = args.slice(2);
    const capturedScopes = evaluate_1.getCapturedScopes(state);
    const fn = (fArgs) => (...aArgs) => {
        return exports.$__scope(state, name, capturedScopes)(false, false, [
            [state.config.reservedNames.self, fn],
            ...(fArgs.map((x, index) => [
                x.symbol,
                ast_1.quote(state, (lastIsSpread && index === fArgs.length - 1) ?
                    aArgs.slice(index) : aArgs[index])
            ])),
        ], ...fnBody);
    };
    const m = {
        name: car.symbol,
        fn: (st, nm, fArgs) => (list) => fn(fArgs)(...(list.slice(1))),
        formalArgs,
        lastIsSpread,
    };
    errors_1.checkUnsafeVarNamesEx('$__defmacro', state.macroMap, car.symbol);
    if (state.macroMap.has(car.symbol)) {
        let curr = state.macroMap.get(car.symbol);
        curr.next = m;
        if (curr && curr.formalArgs) {
            if (curr.formalArgs.length < formalArgs.length) {
                state.macroMap.set(car.symbol, m);
                m.next = curr;
            }
            else {
                let prev = curr;
                curr = curr.next;
                while (curr) {
                    if (curr.formalArgs) {
                        if (curr.formalArgs.length < formalArgs.length) {
                            prev.next = m;
                            m.next = curr;
                            break;
                        }
                    }
                    prev = curr;
                    curr = curr.next;
                }
            }
        }
    }
    else {
        state.macroMap.set(car.symbol, m);
    }
    return fn;
};
exports.$apply = (state, name) => (...args) => {
    // S expression: ($apply fn arg1 ... argN)
    //  -> S expr  : fn'
    errors_1.checkParamsLength('$apply', args, 1);
    const car = exports.$$first(...args);
    if (typeof car !== 'function') {
        throw new Error(`[SX] $apply: Invalid argument(s): args[0] is not function.`);
    }
    return ((...p) => car.apply(null, args.slice(1).concat(p)));
};
exports.$$apply = exports.$apply(null, null);
// tslint:disable-next-line:variable-name
exports.$__call = (state, name) => (...args) => {
    // S expression: ($__call thisArg 'symbol arg1 ... argN)
    //  -> S expr  : fn
    errors_1.checkParamsLength('$__call', args, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    const sym = ast_1.isSymbol(cdr);
    const xName = sym ? sym.symbol : evaluate_1.evaluate(state, cdr);
    errors_1.checkUnsafeVarNamesEx('$__call', car, xName);
    return Function.prototype.apply.call(car[xName], car, args.slice(2));
};
// tslint:disable-next-line:variable-name
exports.$__try = (state, name) => (...args) => {
    // S expression: ($__try 'expr 'catch-expr)
    //  ->                               S expr  : expr
    //  -> (if error is raised in expr)  S expr  : catch-expr
    errors_1.checkParamsLength('$__try', args, 1, 2);
    let r = [];
    try {
        r = evaluate_1.evaluate(state, args[0]);
    }
    catch (e) {
        if (e instanceof types_1.FatalError) {
            throw e;
        }
        if (1 < args.length) {
            r = exports.$__scope(state, name)(true, false, [
                ['$error', ast_1.quote(state, e)],
                ['$parent', ast_1.quote(state, evaluate_1.getScope(state))],
            ], args[1]);
        }
        else {
            r = null;
        }
    }
    return r;
};
exports.$raise = (state, name) => (...args) => {
    // S expression: ($raise 'expr)
    //  -> S expr  : -
    const car = exports.$$first(...args);
    throw car;
};
exports.$$raise = exports.$raise(null, null);
// tslint:disable-next-line:variable-name
exports.$__if = (state, name) => (...args) => {
    // S expression: ($__if condition 't-expr 'f-expr)
    //  -> (if condition is true ) S expr  : t-expr
    //  -> (if condition is false) S expr  : f-expr
    errors_1.checkParamsLength('$__if', args, 2, 3);
    const car = exports.$$first(...args);
    let r = [];
    if (exports.$$boolean(car)) {
        r = evaluate_1.evaluate(state, args[1]);
    }
    else {
        if (2 < args.length) {
            r = evaluate_1.evaluate(state, args[2]);
        }
        else {
            r = null;
        }
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__ifNull = (state, name) => (...args) => {
    // S expression: ($__if-null condition 'null-expr)
    //  -> (if condition is not null ) S expr  : condition
    //  -> (if condition is null)      S expr  : null-expr
    errors_1.checkParamsLength('$__ifNull', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    let r = [];
    if (!exports.$$ambiguousEq(car, null)) {
        r = car;
    }
    else {
        r = evaluate_1.evaluate(state, cdr);
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__cond = (state, name) => (...args) => {
    // S expression: ($__cond 'cond1 'expr1 ... 'condN 'exprN)
    //  -> (if (eval condI) is true ) S expr  : exprI
    //  -> (if no matched)            S expr  : null
    errors_1.checkParamsLength('$__cond', args, 1);
    for (let i = 0; i < args.length - 1; i += 2) {
        const c = args[i];
        const x = args[i + 1];
        if (exports.$$boolean(evaluate_1.evaluate(state, c))) {
            return evaluate_1.evaluate(state, x);
        }
    }
    return null;
};
// tslint:disable-next-line:variable-name
exports.$__while = (state, name) => (...args) => {
    // S expression: ($__while 'condition 'expr1 ... 'exprN)
    //  -> (if condition is true at least 1 or more times) S expr  : exprN
    //  -> (else)                                          S expr  : null
    errors_1.checkParamsLength('$__while', args, 1);
    const car = exports.$$first(...args);
    const cdr = args.slice(1);
    let r = null;
    while (exports.$$boolean(evaluate_1.evaluate(state, car))) {
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__doWhile = (state, name) => (...args) => {
    // S expression: ($__do-while 'condition 'expr1 ... 'exprN)
    //  -> (if condition is true at least 1 or more times) S expr  : exprN
    //  -> (else)                                          S expr  : null
    errors_1.checkParamsLength('$__doWhile', args, 1);
    const car = exports.$$first(...args);
    const cdr = args.slice(1);
    let r = null;
    do {
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    } while (exports.$$boolean(evaluate_1.evaluate(state, car)));
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__until = (state, name) => (...args) => {
    // S expression: ($__until 'condition 'expr1 ... 'exprN)
    //  -> (if condition is true at least 1 or more times) S expr  : exprN
    //  -> (else)                                          S expr  : null
    errors_1.checkParamsLength('$__until', args, 1);
    const car = exports.$$first(...args);
    const cdr = args.slice(1);
    let r = null;
    while (exports.$$not(evaluate_1.evaluate(state, car))) {
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__doUntil = (state, name) => (...args) => {
    // S expression: ($__do-until 'condition 'expr1 ... 'exprN)
    //  -> (if condition is true at least 1 or more times) S expr  : exprN
    //  -> (else)                                          S expr  : null
    errors_1.checkParamsLength('$__doUntil', args, 1);
    const car = exports.$$first(...args);
    const cdr = args.slice(1);
    let r = null;
    do {
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    } while (exports.$$not(evaluate_1.evaluate(state, car)));
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__repeat = (state, name) => (...args) => {
    // S expression: ($__repeat 'i n-times 'expr1 ... 'exprN)
    //  -> (if n > 0) S expr  : exprN
    //  -> (else)     S expr  : null
    errors_1.checkParamsLength('$__repeat', args, 2);
    const sym = ast_1.isSymbol(exports.$$first(...args));
    if (!sym) {
        throw new Error(`[SX] $__repeat: Invalid argument(s): item(s) of args[0] is not symbol.`);
    }
    errors_1.checkUnsafeVarNamesEx('$__repeat', {}, sym.symbol); // NOTE: pass dummy target object.
    const scope = evaluate_1.resolveValueSymbolScope(state, sym, false);
    const n = evaluate_1.toNumber(exports.$$second(...args));
    const cdr = args.slice(2);
    let r = null;
    for (let i = 0; i < n; i++) {
        scope[sym.symbol] = i;
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$__for = (state, name) => (...args) => {
    // S expression: ($__for 'x list 'expr1 ... 'exprN)
    //  -> (if list.length > 0) S expr  : exprN
    //  -> (else)               S expr  : null
    errors_1.checkParamsLength('$__for', args, 2);
    const sym = ast_1.isSymbol(exports.$$first(...args));
    if (!sym) {
        throw new Error(`[SX] $__for: Invalid argument(s): item(s) of args[0] is not symbol.`);
    }
    errors_1.checkUnsafeVarNamesEx('$__for', {}, sym.symbol); // NOTE: pass dummy target object.
    const scope = evaluate_1.resolveValueSymbolScope(state, sym, false);
    const list = exports.$$second(...args);
    if (!Array.isArray(list)) {
        throw new Error(`[SX] $__for: Invalid argument(s): item(s) of args[1] is not array.`);
    }
    const cdr = args.slice(2);
    let r = null;
    for (const q of list) {
        scope[sym.symbol] = q;
        for (const x of cdr) {
            r = evaluate_1.evaluate(state, x);
        }
    }
    return r;
};
// tslint:disable-next-line:variable-name
exports.$pipe = (state, name) => (...args) => {
    // S expression: ($__get v fn1 ... fnN)
    //  -> S expr  : any
    errors_1.checkParamsLength('$pipe', args, 1);
    let v = args[0];
    for (let i = 1; i < args.length; i++) {
        v = args[i](v);
    }
    return v;
};
exports.$$pipe = exports.$pipe(null, null);
// tslint:disable-next-line:variable-name
exports.$__get = (state, name) => (...args) => {
    // S expression: ($__get 'nameOrIndex1 ... 'nameOrIndexN)
    //  -> S expr  : any
    errors_1.checkParamsLength('$__get', args, 1);
    let v = null;
    let sym = ast_1.isSymbol(args[0]);
    if (!sym) {
        switch (typeof args[0]) {
            case 'string':
            case 'number':
                sym = { symbol: String(args[0]) };
                break;
            default:
                v = evaluate_1.evaluate(state, args[0]);
                break;
        }
    }
    if (sym) {
        const scope = evaluate_1.resolveValueSymbolScope(state, sym, true);
        if (!scope) {
            throw new Error(`[SX] $__get: Invalid argument(s): args[0]: symbol "${sym.symbol}" is not defined.`);
        }
        v = scope[sym.symbol];
    }
    for (let i = 1; i < args.length; i++) {
        let q = args[i];
        let inprog = true;
        while (inprog) {
            switch (typeof q) {
                case 'function':
                    v = q(v);
                    inprog = false;
                    break;
                case 'object':
                    if (Array.isArray(q)) {
                        q = evaluate_1.evaluate(state, q);
                    }
                    else {
                        sym = ast_1.isSymbol(q);
                        if (sym) {
                            q = sym.symbol;
                        }
                        else if (Object.prototype.hasOwnProperty.call(q, 'value')) {
                            q = evaluate_1.evaluate(state, q);
                        }
                        else {
                            throw new Error(`[SX] $__get: Invalid argument(s): invalid name path.`);
                        }
                    }
                    break;
                case 'number':
                    if (q < 0) {
                        q = v.length + q;
                    }
                // FALL_THRU
                case 'string':
                    errors_1.checkUnsafeVarNamesEx('$__get', v, q);
                    v = v[q];
                    inprog = false;
                    break;
                default:
                    throw new Error(`[SX] $__get: Invalid argument(s): invalid name path.`);
            }
        }
    }
    return v;
};
// tslint:disable-next-line:variable-name
exports.$__let = (state, name) => (...args) => {
    // S expression: ($__let 'nameStrOrSymbol expr)
    //  -> S expr  : any
    errors_1.checkParamsLength('$__let', args, 2, 2);
    let sym = ast_1.isSymbol(exports.$$first(...args));
    if (!sym) {
        if (typeof args[0] === 'string') {
            sym = { symbol: args[0] };
        }
        else {
            throw new Error(`[SX] $__let: Invalid argument(s): invalid name.`);
        }
    }
    errors_1.checkUnsafeVarNamesEx('$__let', {}, sym.symbol); // NOTE: pass dummy target object.
    const scope = evaluate_1.resolveValueSymbolScope(state, sym, false);
    scope[sym.symbol] = args[1];
    return args[1];
};
// tslint:disable-next-line:variable-name
exports.$__set = (state, name) => (...args) => {
    // S expression: ($__set 'nameOrListOfNameOrIndex expr)
    //  -> S expr  : any
    errors_1.checkParamsLength('$__set', args, 2, 2);
    let path = [];
    if (Array.isArray(args[0])) {
        path = args[0];
    }
    else {
        path.push(args[0]);
    }
    let sym = ast_1.isSymbol(path[0]);
    if (!sym) {
        if (typeof path[0] === 'string') {
            errors_1.checkUnsafeVarNamesEx('$__set', {}, path[0]); // NOTE: pass dummy target object.
            sym = { symbol: path[0] };
        }
        else {
            throw new Error(`[SX] $__set: Invalid argument(s): invalid name.`);
        }
    }
    let scope = evaluate_1.resolveValueSymbolScope(state, sym, true);
    if (scope === null) {
        throw new Error(`[SX] $__set: Unresolved symbol: ${sym.symbol}.`);
    }
    let subst = false;
    for (let i = 0; i < path.length; i++) {
        let q = path[i];
        let inprog = true;
        const last = i === path.length - 1;
        while (inprog) {
            switch (typeof q) {
                case 'function':
                    scope = q(scope);
                    inprog = false;
                    break;
                case 'object':
                    if (Array.isArray(q)) {
                        q = evaluate_1.evaluate(state, q);
                    }
                    else {
                        sym = ast_1.isSymbol(q);
                        if (sym) {
                            q = sym.symbol;
                        }
                        else if (Object.prototype.hasOwnProperty.call(q, 'value')) {
                            q = evaluate_1.evaluate(state, q);
                        }
                        else {
                            throw new Error(`[SX] $__set: Invalid argument(s): invalid name.`);
                        }
                    }
                    break;
                case 'number':
                    if (q < 0) {
                        q = scope.length + q;
                    }
                // FALL_THRU
                case 'string':
                    errors_1.checkUnsafeVarNamesEx('$__set', scope, q);
                    if (last) {
                        scope[q] = args[1];
                        subst = true;
                    }
                    else {
                        scope = scope[q];
                    }
                    inprog = false;
                    break;
                default:
                    throw new Error(`[SX] $__set: Invalid argument(s): invalid name.`);
            }
        }
    }
    if (!subst) {
        throw new Error(`[SX] $__set: Invalid argument(s): last path is not lvalue.`);
    }
    return args[1];
};
exports.$boolean = (state, name) => (...args) => {
    // S expression: ($boolean any)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$boolean', args, 1, 1);
    const car = exports.$$first(...args);
    if (Array.isArray(car) && car.length === 0)
        return false;
    else
        return Boolean(car);
};
exports.$$boolean = exports.$boolean(null, null);
exports.$not = (state, name) => (...args) => {
    // S expression: ($not any)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$not', args, 1, 1);
    return !exports.$$boolean(...args);
};
exports.$$not = exports.$not(null, null);
// tslint:disable-next-line:variable-name
exports.$__and = (state, name) => (...args) => {
    // S expression: ($__and 'expr1 ... 'exprN)
    //  -> (if all of ($boolean expr1) ... ($boolean exprN) are true) S expr  : exprN
    //  -> (else)                                                     S expr  : expr-i (false left most)
    errors_1.checkParamsLength('$__and', args, 1);
    let prev = null;
    for (let i = 0; i < args.length; i++) {
        const curr = evaluate_1.evaluate(state, args[i]);
        if (!exports.$$boolean(curr)) {
            return curr;
        }
        prev = curr;
    }
    return prev;
};
// tslint:disable-next-line:variable-name
exports.$$__and = exports.$__and(null, null);
// tslint:disable-next-line:variable-name
exports.$__or = (state, name) => (...args) => {
    // S expression: ($__or 'expr1 ... 'exprN)
    //  -> (if any ($boolean expr1) ... ($boolean exprN) are true) S expr  : expr-i (where i: index of item first ($boolean expr-i) is to be true)
    //  -> (else)                                                  S expr  : expr-i (false right most)
    errors_1.checkParamsLength('$__or', args, 1);
    let prev = null;
    for (let i = 0; i < args.length; i++) {
        const curr = evaluate_1.evaluate(state, args[i]);
        if (exports.$$boolean(curr)) {
            return curr;
        }
        prev = curr;
    }
    return prev;
};
// tslint:disable-next-line:variable-name
exports.$$__or = exports.$__or(null, null);
exports.$ambiguousEq = (state, name) => (...args) => {
    // S expression: (== a b)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$ambiguousEq', args, 2, 2);
    let { car, cdr } = exports.$$firstAndSecond(...args);
    if (Array.isArray(car) && car.length === 0)
        car = null;
    if (Array.isArray(cdr) && cdr.length === 0)
        cdr = null;
    if (car === void 0)
        car = null;
    if (cdr === void 0)
        cdr = null;
    // tslint:disable-next-line:triple-equals
    return car == cdr;
};
exports.$$ambiguousEq = exports.$ambiguousEq(null, null);
exports.$ambiguousNotEq = (state, name) => (...args) => {
    // S expression: (!= a b)
    //  -> S expr  : boolean
    return !exports.$$ambiguousEq(...args);
};
exports.$$ambiguousNotEq = exports.$ambiguousNotEq(null, null);
exports.$lt = (state, name) => (...args) => {
    // S expression: (< a b)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$lt', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return evaluate_1.toNumber(car) < evaluate_1.toNumber(cdr);
};
exports.$$lt = exports.$lt(null, null);
exports.$le = (state, name) => (...args) => {
    // S expression: (<= a b)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$le', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return evaluate_1.toNumber(car) <= evaluate_1.toNumber(cdr);
};
exports.$$le = exports.$le(null, null);
exports.$gt = (state, name) => (...args) => {
    // S expression: (> a b)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$gt', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return evaluate_1.toNumber(car) > evaluate_1.toNumber(cdr);
};
exports.$$gt = exports.$gt(null, null);
exports.$ge = (state, name) => (...args) => {
    // S expression: (>= a b)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$ge', args, 2, 2);
    const { car, cdr } = exports.$$firstAndSecond(...args);
    return evaluate_1.toNumber(car) >= evaluate_1.toNumber(cdr);
};
exports.$$ge = exports.$ge(null, null);
exports.$symbol = (state, name) => (...args) => {
    // S expression: ($symbol)
    // S expression: ($symbol name)
    //  -> S expr  : symbol
    errors_1.checkParamsLength('$symbol', args, 1, 1);
    if (typeof args[0] === 'string') {
        return { symbol: args[0] };
    }
    else {
        throw new Error(`[SX] $symbol: Invalid argument(s): item(s) of args[0] is not string.`);
    }
};
exports.$$symbol = exports.$symbol(null, null);
// tslint:disable-next-line:variable-name
exports.$__gensym = (state, name) => (...args) => {
    // S expression: ($__gensym)
    // S expression: ($__gensym name)
    //  -> S expr  : symbol
    errors_1.checkParamsLength('$__gensym', args, 0, 1);
    const varBaseName = `$__tempvar__$$ec${state.evalCount++}$$_`;
    const tempVarSym = ({ symbol: `${varBaseName}_$gensym` });
    if (args.length === 1) {
        const a = ast_1.isSymbol(args[0]);
        if (a) {
            exports.$__let(state, '')(a, tempVarSym);
        }
        else if (typeof args[0] === 'string') {
            exports.$__let(state, '')({ symbol: args[0] }, tempVarSym);
        }
        else {
            throw new Error(`[SX] $__gensym: Invalid argument(s): item(s) of args[0] is not symbol.`);
        }
    }
    return tempVarSym;
};
exports.$isSymbol = (state, name) => (...args) => {
    // S expression: ($is-symbol x)
    // S expression: ($is-symbol x name)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isSymbol', args, 1, 2);
    if (args.length === 1) {
        return (ast_1.isSymbol(args[0]) ? true : false);
    }
    else {
        if (typeof args[1] === 'string') {
            return (ast_1.isSymbol(args[0], args[1]) ? true : false);
        }
        else {
            throw new Error(`[SX] $isSymbol: Invalid argument(s): item(s) of args[1] is not string.`);
        }
    }
};
exports.$$isSymbol = exports.$isSymbol(null, null);
exports.$isList = (state, name) => (...args) => {
    // S expression: ($is-list x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isList', args, 1, 1);
    return Array.isArray(exports.$$first(...args));
};
exports.$$isList = exports.$isList(null, null);
exports.$isString = (state, name) => (...args) => {
    // S expression: ($is-string x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isString', args, 1, 1);
    return typeof exports.$$first(...args) === 'string';
};
exports.$$isString = exports.$isString(null, null);
exports.$isNumber = (state, name) => (...args) => {
    // S expression: ($is-number x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isNumber', args, 1, 1);
    return typeof exports.$$first(...args) === 'number';
};
exports.$$isNumber = exports.$isNumber(null, null);
exports.$isNaN = (state, name) => (...args) => {
    // S expression: ($is-NaN x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isNaN', args, 1, 1);
    return Number.isNaN(exports.$$first(...args));
};
exports.$$isNaN = exports.$isNaN(null, null);
exports.$isFinite = (state, name) => (...args) => {
    // S expression: ($is-finate x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isFinite', args, 1, 1);
    return Number.isFinite(exports.$$first(...args));
};
exports.$$isFinite = exports.$isFinite(null, null);
exports.$isInteger = (state, name) => (...args) => {
    // S expression: ($is-integer x)
    //  -> S expr  : boolean
    errors_1.checkParamsLength('$isInteger', args, 1, 1);
    return Number.isInteger(exports.$$first(...args));
};
exports.$$isInteger = exports.$isInteger(null, null);
exports.$toString = (state, name) => (...args) => {
    // S expression: ($to-string x)
    //  -> S expr  : string
    errors_1.checkParamsLength('$toString', args, 1, 1);
    return String(exports.$$first(...args));
};
exports.$$toString = exports.$toString(null, null);
exports.$toNumber = (state, name) => (...args) => {
    // S expression: ($to-number x)
    //  -> S expr  : number
    errors_1.checkParamsLength('$toNumber', args, 1, 1);
    return evaluate_1.toNumber(exports.$$first(...args));
};
exports.$$toNumber = exports.$toNumber(null, null);
// tslint:disable-next-line:variable-name
exports.$__toObject = (state, name) => (...args) => {
    // S expression: ($__# '(name value...)...)
    //  -> JSON    : {name: value, ...}
    const r = {};
    for (const x of args) {
        if (Array.isArray(x) && 0 < x.length) {
            const sym = ast_1.isSymbol(x[0]);
            const keyName = sym ? sym.symbol :
                String(evaluate_1.evaluate(state, x[0]));
            errors_1.checkUnsafeVarNamesEx('$__#', r, keyName);
            if (x.length === 1) {
                // S expression: (# ... (keyName) ...)
                //  -> JSON    : {..., keyName: true, ...}
                r[keyName] = true;
            }
            else if (x.length === 2) {
                // S expression: (# ... (keyName value) ...)
                //  -> JSON    : {..., keyName: value, ...}
                r[keyName] = evaluate_1.evaluate(state, x[1]);
            }
            else {
                // S expression: (# ... (keyName value1 value2 ...) ...)
                //  -> JSON    : {..., keyName: [value1, value2, ], ...}
                r[keyName] =
                    evaluate_1.evaluate(state, [{ symbol: state.config.reservedNames.list }]
                        .concat(x.slice(1)));
            }
        }
        else {
            throw new Error(`[SX] $__toObject: Invalid argument(s): args[?] is not array.`);
        }
    }
    return r;
};
const assignBlacklists = [
    consts_1.globalObj,
    Object.__proto__,
    {}.__proto__,
    Function.__proto__,
];
exports.$objectAssign = (state, name) => (...args) => {
    // S expression: ($object-assign x)
    //  -> S expr  : string
    errors_1.checkParamsLength('$objectAssign', args, 1);
    if (assignBlacklists.includes(args[0])) {
        throw new Error(`[SX] $objectAssign: Invalid argument: args[0] is blacklisted object.`);
    }
    return Object.assign(args[0], ...(args.slice(1)));
};
exports.$$objectAssign = exports.$objectAssign(null, null);
exports.$jsonStringify = (state, name) => (...args) => {
    // S expression: ($json-stringify x)
    //  -> S expr  : string
    errors_1.checkParamsLength('$jsonStringify', args, 1, 1);
    return JSON.stringify(exports.$$first(...args));
};
exports.$$jsonStringify = exports.$jsonStringify(null, null);
exports.$jsonParse = (state, name) => (...args) => {
    // S expression: ($json-parse x)
    //  -> S expr  : object
    errors_1.checkParamsLength('$jsonParse', args, 1, 1);
    const s = exports.$$first(...args);
    if (typeof s !== 'string') {
        throw new Error(`[SX] $jsonParse: Invalid argument(s): args[0] is not string.`);
    }
    return JSON.parse(s);
};
exports.$$jsonParse = exports.$jsonParse(null, null);
exports.$now = (state, name) => (...args) => {
    // S expression: ($now)
    //  -> S expr  : number
    return Date.now();
};
exports.$$now = exports.$now(null, null);
exports.$datetimeFromIso = (state, name) => (...args) => {
    // S expression: ($datetime-from-iso str)
    //  -> S expr  : number
    errors_1.checkParamsLength('$datetimeFromIso', args, 1, 1);
    const s = exports.$$first(...args);
    if (typeof s !== 'string') {
        throw new Error(`[SX] $datetimeFromIso: Invalid argument(s): args[0] is not string.`);
    }
    if (!/^(?:(?:-[0-9]{6,})|[0-9]{4,})-(?:[0-1][0-9])-(?:[0-3][0-9])(?:T(?:[0-2][0-9])(?:[:](?:[0-6][0-9])(?:[:](?:[0-6][0-9])(?:.[0-9]{1,})?)?)?(?:Z|[-+][0-9]{2}(?:[:]?[0-6][0-9])?)?)?$/.test(s)) {
        throw new Error(`[SX] $datetimeFromIso: Invalid datetime (pattern unmatched): ${s}.`);
    }
    const dt = new Date(s).getTime();
    if (Number.isNaN(dt)) {
        throw new Error(`[SX] $datetimeFromIso: Invalid datetime: ${s}.`);
    }
    return dt;
};
exports.$$datetimeFromIso = exports.$datetimeFromIso(null, null);
exports.$datetime = (state, name) => (...args) => {
    // S expression: ($datetime year month1-12 day)
    // S expression: ($datetime year month1-12 day hours)
    // S expression: ($datetime year month1-12 day hours minutes)
    // S expression: ($datetime year month1-12 day hours minutes seconds)
    // S expression: ($datetime year month1-12 day hours minutes seconds milliseconds)
    //  -> S expr  : number
    errors_1.checkParamsLength('$datetime', args, 3, 7);
    let s = '';
    const year = Number(args[0]);
    if (year >= 0) {
        s += String(year).padStart(4, '0');
    }
    else {
        s += '-' + String(-year).padStart(6, '0');
    }
    // month1
    s += '-' + String(Number(args[1])).padStart(2, '0');
    // day
    s += '-' + String(Number(args[2])).padStart(2, '0');
    // hours
    if (args.length >= 4) {
        s += 'T' + String(Number(args[3])).padStart(2, '0');
        // minutes
        if (args.length >= 5) {
            s += ':' + String(Number(args[4])).padStart(2, '0');
        }
        else {
            s += ':00';
        }
        // seconds
        if (args.length >= 6) {
            s += ':' + String(Number(args[5])).padStart(2, '0');
        }
        // milliseconds
        if (args.length >= 7) {
            s += '.' + String(Number(args[6])).padStart(3, '0').slice(0, 3);
        }
        s += 'Z';
    }
    const dt = new Date(s).getTime();
    if (Number.isNaN(dt)) {
        throw new Error(`[SX] $datetime: Invalid datetime: ${s}.`);
    }
    return dt;
};
exports.$$datetime = exports.$datetime(null, null);
exports.$datetimeLc = (state, name) => (...args) => {
    // S expression: ($datetime-lc year month1-12 day)
    // S expression: ($datetime-lc year month1-12 day hours)
    // S expression: ($datetime-lc year month1-12 day hours minutes)
    // S expression: ($datetime-lc year month1-12 day hours minutes seconds)
    // S expression: ($datetime-lc year month1-12 day hours minutes seconds milliseconds)
    //  -> S expr  : number
    errors_1.checkParamsLength('$datetimeLc', args, 3, 7);
    let s = '';
    const year = Number(args[0]);
    if (year >= 0) {
        s += String(year).padStart(4, '0');
    }
    else {
        s += '-' + String(-year).padStart(6, '0');
    }
    // month1
    s += '-' + String(Number(args[1])).padStart(2, '0');
    // day
    s += '-' + String(Number(args[2])).padStart(2, '0');
    // hours
    if (args.length >= 4) {
        s += 'T' + String(Number(args[3])).padStart(2, '0');
        // minutes
        if (args.length >= 5) {
            s += ':' + String(Number(args[4])).padStart(2, '0');
        }
        else {
            s += ':00';
        }
        // seconds
        if (args.length >= 6) {
            s += ':' + String(Number(args[5])).padStart(2, '0');
        }
        // milliseconds
        if (args.length >= 7) {
            s += '.' + String(Number(args[6])).padStart(3, '0').slice(0, 3);
        }
    }
    else {
        s += 'T00:00:00.000';
    }
    const dt = new Date(s).getTime();
    if (Number.isNaN(dt)) {
        throw new Error(`[SX] $datetimeLc: Invalid datetime: ${s}.`);
    }
    return dt;
};
exports.$$datetimeLc = exports.$datetimeLc(null, null);
exports.$datetimeToIsoString = (state, name) => (...args) => {
    // S expression: ($datetime-to-iso-string number)
    //  -> S expr  : string
    errors_1.checkParamsLength('$datetimeToIsoString', args, 1, 1);
    const n = exports.$$first(...args);
    if (typeof n !== 'number') {
        throw new Error(`[SX] $datetimeToIsoString: Invalid argument(s): args[0] is not number.`);
    }
    const dt = new Date(n);
    if (Number.isNaN(dt.getTime())) {
        throw new Error(`[SX] $datetimeToIsoString: Invalid datetime: ${n}.`);
    }
    return dt.toISOString();
};
exports.$$datetimeToIsoString = exports.$datetimeToIsoString(null, null);
exports.$datetimeToComponents = (state, name) => (...args) => {
    // S expression: ($datetime-to-components number)
    //  -> S expr  : list
    errors_1.checkParamsLength('$datetimeToComponents', args, 1, 1);
    const n = exports.$$first(...args);
    if (typeof n !== 'number') {
        throw new Error(`[SX] $datetimeToComponents: Invalid argument(s): args[0] is not number.`);
    }
    const dt = new Date(n);
    if (Number.isNaN(dt.getTime())) {
        throw new Error(`[SX] $datetimeToComponents: Invalid datetime: ${n}.`);
    }
    return ([
        dt.getUTCFullYear(),
        dt.getUTCMonth() + 1,
        dt.getUTCDate(),
        dt.getUTCHours(),
        dt.getUTCMinutes(),
        dt.getUTCSeconds(),
        dt.getUTCMilliseconds(),
        0,
        dt.getUTCDay(),
    ]);
};
exports.$$datetimeToComponents = exports.$datetimeToComponents(null, null);
exports.$datetimeToComponentsLc = (state, name) => (...args) => {
    // S expression: ($datetime-to-components-lc number)
    //  -> S expr  : list
    errors_1.checkParamsLength('$datetimeToComponentsLc', args, 1, 1);
    const n = exports.$$first(...args);
    if (typeof n !== 'number') {
        throw new Error(`[SX] $datetimeToComponentsLc: Invalid argument(s): args[0] is not number.`);
    }
    const dt = new Date(n);
    if (Number.isNaN(dt.getTime())) {
        throw new Error(`[SX] $datetimeToComponentsLc: Invalid datetime: ${n}.`);
    }
    return ([
        dt.getFullYear(),
        dt.getMonth() + 1,
        dt.getDate(),
        dt.getHours(),
        dt.getMinutes(),
        dt.getSeconds(),
        dt.getMilliseconds(),
        -dt.getTimezoneOffset(),
        // If your time zone is UTC+2:00, +120 will be returned.
        dt.getDay(),
    ]);
};
exports.$$datetimeToComponentsLc = exports.$datetimeToComponentsLc(null, null);
exports.$match = (state, name) => (...args) => {
    // S expression: ($match pattern-str string)
    // S expression: ($match pattern-str options-str string)
    //  -> S expr  : array
    errors_1.checkParamsLength('$match', args, 2, 3);
    if (!state.config.enableRegExpMatchOperators) {
        throw new Error(`[SX] $match: Operator is disabled by configuration.`);
    }
    if (args.length === 2) {
        const m = new RegExp(args[0]);
        return m.exec(args[1]);
    }
    else {
        const m = new RegExp(args[0], args[1]);
        return m.exec(args[2]);
    }
};
exports.$$match = exports.$match(null, null);
exports.$consoleLog = (state, name) => (...args) => {
    // S expression: ($console-log expr1 ... exprN)
    //  -> S expr  : null
    console.log(...args);
    return null;
};
exports.$$consoleLog = exports.$consoleLog(null, null);
exports.$consoleError = (state, name) => (...args) => {
    // S expression: ($console-error expr1 ... exprN)
    //  -> S expr  : null
    console.error(...args);
    return null;
};
exports.$$consoleError = exports.$consoleError(null, null);
exports.$consoleTrace = (state, name) => (...args) => {
    // S expression: ($console-trace expr1 ... exprN)
    //  -> S expr  : null
    console.trace(...args);
    return null;
};
exports.$$consoleTrace = exports.$consoleTrace(null, null);
exports.$consoleTime = (state, name) => (...args) => {
    // S expression: ($console-time)
    // S expression: ($console-time label)
    //  -> S expr  : null
    console.time(...args);
    return null;
};
exports.$$consoleTime = exports.$consoleTime(null, null);
exports.$consoleTimeEnd = (state, name) => (...args) => {
    // S expression: ($console-time-end)
    // S expression: ($console-time-end label)
    //  -> S expr  : null
    console.timeEnd(...args);
    return null;
};
exports.$$consoleTimeEnd = exports.$consoleTimeEnd(null, null);
exports.$consoleTimeLog = (state, name) => (...args) => {
    // S expression: ($console-time-log label)
    // S expression: ($console-time-log label value ... value)
    //  -> S expr  : null
    console.timeLog(...args);
    return null;
};
exports.$$consoleTimeLog = exports.$consoleTimeLog(null, null);
//# sourceMappingURL=core.fn.js.map