"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const evaluate_1 = require("../../evaluate");
const errors_1 = require("../../errors");
const core_fn_1 = require("../core/core.fn");
const data_1 = require("../../../lib/data");
exports.$range = (state, name) => (...args) => {
    // S expression: ($range start end)
    // S expression: ($range start end step)
    //  -> S expr  : list
    errors_1.checkParamsLength('$range', args, 2, 3);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    const start = evaluate_1.toNumber(car) || 0;
    const stop = evaluate_1.toNumber(cdr) || 0;
    const step = (args.length > 2 ? evaluate_1.toNumber(args[2]) || 0 : 0) || (start <= stop ? 1 : -1);
    const n = Math.sign(stop - start) + Math.sign(step) !== 0 ?
        (Math.floor((Math.abs(stop - start) / Math.abs(step))) + 1) : 0;
    state.evalCount += n;
    evaluate_1.evaluate(state, 0);
    return Array.from({ length: n }, (x, i) => start + i * step);
};
exports.$length = (state, name) => (...args) => {
    // S expression: ($length listOrString)
    //  -> S expr  : number
    errors_1.checkParamsLength('$length', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    switch (typeof car) {
        case 'object':
            if (!('length' in car)) {
                break;
            }
        // FALL_THRU
        case 'string':
            return car.length;
    }
    throw new Error(`[SX] $length: Invalid argument type: object has no property 'length'.`);
};
exports.$$length = exports.$length(null, null);
exports.$trim = (state, name) => (...args) => {
    // S expression: ($trim string)
    //  -> S expr  : string
    errors_1.checkParamsLength('$trim', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    if (typeof car === 'string') {
        return car.trim();
    }
    throw new Error(`[SX] $trim: Invalid argument type: args[0] is not string.`);
};
exports.$$trim = exports.$trim(null, null);
exports.$trimHead = (state, name) => (...args) => {
    // S expression: ($trim-head string)
    //  -> S expr  : string
    errors_1.checkParamsLength('$trimHead', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    if (typeof car === 'string') {
        return car.trimLeft();
    }
    throw new Error(`[SX] $trimHead: Invalid argument type: args[0] is not string.`);
};
exports.$$trimHead = exports.$trimHead(null, null);
exports.$trimTail = (state, name) => (...args) => {
    // S expression: ($trim-tail string)
    //  -> S expr  : string
    errors_1.checkParamsLength('$trimTail', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    if (typeof car === 'string') {
        return car.trimRight();
    }
    throw new Error(`[SX] $trimTail: Invalid argument type: args[0] is not string.`);
};
exports.$$trimTail = exports.$trimTail(null, null);
exports.$replaceAll = (state, name) => (...args) => {
    // S expression: ($replace-all src-string match-string replacement-string)
    //  -> S expr  : string
    errors_1.checkParamsLength('$replaceAll', args, 3, 3);
    if (typeof args[0] === 'string' && typeof args[1] === 'string' && typeof args[2] === 'string') {
        return args[0].split(args[1]).join(args[2]);
    }
    throw new Error(`[SX] $replaceAll: Invalid argument type: args[0] or [1] or [2] is not string.`);
};
exports.$$replaceAll = exports.$replaceAll(null, null);
exports.$split = (state, name) => (...args) => {
    // S expression: ($split src-string match-string)
    //  -> S expr  : (string ... string)
    errors_1.checkParamsLength('$split', args, 2, 2);
    if (typeof args[0] === 'string' && typeof args[1] === 'string') {
        return args[0].split(args[1]);
    }
    throw new Error(`[SX] $split: Invalid argument type: args[0] or [1] is not string.`);
};
exports.$$split = exports.$split(null, null);
exports.$join = (state, name) => (...args) => {
    // S expression: ($join '(str1 ... strN) separator)
    //  -> S expr  : (string ... string)
    errors_1.checkParamsLength('$join', args, 1, 2);
    if (typeof Array.isArray(args[0])) {
        if (args.length > 1) {
            if (typeof args[1] === 'string') {
                return args[0].join(args[1]);
            }
            throw new Error(`[SX] $join: Invalid argument type: args[1] is not string.`);
        }
        else {
            return args[0].join();
        }
    }
    throw new Error(`[SX] $join: Invalid argument type: args[0] is not array.`);
};
exports.$$join = exports.$join(null, null);
exports.$concat = (state, name) => (...args) => {
    // S expression: ($concat listOrString1 ... listOrStringN)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$concat', args, 1);
    const car = core_fn_1.$$first(...args);
    switch (typeof car) {
        case 'object':
            if (!('concat' in car)) {
                break;
            }
        // FALL_THRU
        case 'string':
            return car.concat(...args.slice(1));
    }
    throw new Error(`[SX] $concat: Invalid argument type: object has no property 'concat'.`);
};
exports.$$concat = exports.$concat(null, null);
exports.$slice = (state, name) => (...args) => {
    // S expression: ($slice start end listOrString)
    // S expression: ($slice start listOrString)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$slice', args, 2, 3);
    if (args.length === 3) {
        if (typeof args[2] === 'string' || Array.isArray(args[2])) {
            return args[2].slice(evaluate_1.toNumber(args[0]), evaluate_1.toNumber(args[1]));
        }
    }
    if (args.length === 2) {
        if (typeof args[1] === 'string' || Array.isArray(args[1])) {
            return args[1].slice(evaluate_1.toNumber(args[0]));
        }
    }
    throw new Error(`[SX] $slice: Invalid argument type: args[${args.length - 1}] is not string or array.`);
};
exports.$$slice = exports.$slice(null, null);
exports.$top = (state, name) => (...args) => {
    // S expression: ($top n listOrString)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$top', args, 2, 2);
    if (typeof args[1] === 'string' || Array.isArray(args[1])) {
        return args[1].slice(0, evaluate_1.toNumber(args[0]));
    }
    throw new Error(`[SX] $top: Invalid argument type: args[1] is not string or array.`);
};
exports.$$top = exports.$top(null, null);
exports.$tail = (state, name) => (...args) => {
    // S expression: ($tail n listOrString)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$tail', args, 2, 2);
    if (typeof args[1] === 'string' || Array.isArray(args[1])) {
        const n = -evaluate_1.toNumber(args[0]);
        return args[1].slice(n >= 0 || Number.isNaN(n) ? args[1].length : n);
    }
    throw new Error(`[SX] $tail: Invalid argument type: args[1] is not string or array.`);
};
exports.$$tail = exports.$tail(null, null);
exports.$push = (state, name) => (...args) => {
    // S expression: ($push list value)
    //  -> S expr  : list
    errors_1.checkParamsLength('$push', args, 2, 2);
    if (typeof Array.isArray(args[0])) {
        args[0].push(args[1]);
        return args[0];
    }
    throw new Error(`[SX] $push: Invalid argument type: args[1] is not array.`);
};
exports.$$push = exports.$push(null, null);
exports.$pop = (state, name) => (...args) => {
    // S expression: ($pop list)
    //  -> S expr  : value
    errors_1.checkParamsLength('$pop', args, 1, 1);
    if (typeof Array.isArray(args[0])) {
        const v = args[0].pop();
        return v;
    }
    throw new Error(`[SX] $pop: Invalid argument type: args[1] is not array.`);
};
exports.$$pop = exports.$pop(null, null);
// tslint:disable-next-line:variable-name
exports.$__at = (state, name) => (...args) => {
    // S expression: ($__at index listOrString)
    //  -> S expr  : any
    errors_1.checkParamsLength('$__at', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    return cdr[car];
};
// tslint:disable-next-line:variable-name
exports.$$__at = exports.$__at(null, null);
exports.$reverse = (state, name) => (...args) => {
    // S expression: ($reverse listOrString)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$reverse', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    if (Array.isArray(car)) {
        return car.slice(0).reverse();
    }
    throw new Error(`[SX] $reverse: Invalid argument type: args[0] is not array.`);
};
exports.$$reverse = exports.$reverse(null, null);
exports.$reverseDestructive = (state, name) => (...args) => {
    // S expression: ($reverse! listOrString)
    //  -> S expr  : listOrString
    errors_1.checkParamsLength('$reverse!', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    if (Array.isArray(car)) {
        return car.reverse();
    }
    throw new Error(`[SX] $reverse!: Invalid argument type: args[0] is not array.`);
};
exports.$$reverseDestructive = exports.$reverseDestructive(null, null);
exports.$find = (state, name) => (...args) => {
    // S expression: ($find list (lambda (v index array) (... boolean)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$find', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        return car.find(cdr);
    }
    throw new Error(`[SX] $find: Invalid argument type: args[0] is not array.`);
};
exports.$$find = exports.$find(null, null);
exports.$filter = (state, name) => (...args) => {
    // S expression: ($filter list (lambda (v index array) (... boolean)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$filter', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        return car.filter(cdr);
    }
    throw new Error(`[SX] $filter: Invalid argument type: args[0] is not array.`);
};
exports.$$filter = exports.$filter(null, null);
exports.$map = (state, name) => (...args) => {
    // S expression: ($map list (lambda (v index array) (... any)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$map', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        return car.map(cdr);
    }
    throw new Error(`[SX] $map: Invalid argument type: args[0] is not array.`);
};
exports.$$map = exports.$map(null, null);
exports.$reduce = (state, name) => (...args) => {
    // S expression: ($reduce list (lambda (acc v index array) (... any)) initial-value)
    // S expression: ($reduce list (lambda (acc v index array) (... any)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$reduce', args, 2, 3);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        if (args.length < 3) {
            return car.reduce(cdr);
        }
        else {
            return car.reduce(cdr, args[2]);
        }
    }
    throw new Error(`[SX] $reduce: Invalid argument type: args[0] is not array.`);
};
exports.$$reduce = exports.$reduce(null, null);
exports.$reduceFromTail = (state, name) => (...args) => {
    // S expression: ($reduce-from-tail list (lambda (acc v index array) (... any)) initial-value)
    // S expression: ($reduce-from-tail list (lambda (acc v index array) (... any)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$reduceFromTail', args, 2, 3);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        if (args.length < 3) {
            return car.reduceRight(cdr);
        }
        else {
            return car.reduceRight(cdr, args[2]);
        }
    }
    throw new Error(`[SX] $reduceFromTail: Invalid argument type: args[0] is not array.`);
};
exports.$$reduceFromTail = exports.$reduceFromTail(null, null);
exports.$sort = (state, name) => (...args) => {
    // S expression: ($sort list (lambda (a b) (... number_a-b)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$sort', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        return car.slice(0).sort(cdr);
    }
    throw new Error(`[SX] $sort: Invalid argument type: args[0] is not array.`);
};
exports.$$sort = exports.$sort(null, null);
exports.$sortDestructive = (state, name) => (...args) => {
    // S expression: ($sort! list (lambda (a b) (... number_a-b)))
    //  -> S expr  : list
    errors_1.checkParamsLength('$sort!', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (Array.isArray(car)) {
        return car.sort(cdr);
    }
    throw new Error(`[SX] $sort!: Invalid argument type: args[0] is not array.`);
};
exports.$$sortDestructive = exports.$sortDestructive(null, null);
exports.$groupEvery = (state, name) => (...args) => {
    // S expression: ($group-every optionsOrNumber (x1 ... xN))
    //  -> S expr  : ((x1 ... ) ... ( ... xN))
    errors_1.checkParamsLength('$group-every', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (!Array.isArray(cdr)) {
        throw new Error(`[SX] $group-every: Invalid argument type: args[1] is not array.`);
    }
    return data_1.query(cdr).groupEvery(car).select();
};
exports.$$groupEvery = exports.$groupEvery(null, null);
exports.$groupBy = (state, name) => (...args) => {
    // S expression: ($group-by conditions (x1 ... xN))
    //  -> S expr  : ((x1 ... ) ... ( ... xN))
    errors_1.checkParamsLength('$group-by', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (!Array.isArray(cdr)) {
        throw new Error(`[SX] $group-by: Invalid argument type: args[1] is not array.`);
    }
    return data_1.query(cdr).groupBy(car).select();
};
exports.$$groupBy = exports.$groupBy(null, null);
exports.$orderBy = (state, name) => (...args) => {
    // S expression: ($order-by conditions (x1 ... xN))
    //  -> S expr  : (x1 ... xN)
    errors_1.checkParamsLength('$order-by', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (!Array.isArray(cdr)) {
        throw new Error(`[SX] $order-by: Invalid argument type: args[1] is not array.`);
    }
    return data_1.query(cdr).orderBy(car).select();
};
exports.$$orderBy = exports.$orderBy(null, null);
exports.$where = (state, name) => (...args) => {
    // S expression: ($where (-> (v index array) ... boolean) (x1 ... xN))
    //  -> S expr  : (x'1 ... x'M)
    errors_1.checkParamsLength('$where', args, 2, 2);
    const { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    if (typeof args[0] !== 'function') {
        throw new Error(`[SX] $where: Invalid argument type: args[0] is not function.`);
    }
    if (!Array.isArray(cdr)) {
        throw new Error(`[SX] $where: Invalid argument type: args[1] is not array.`);
    }
    return data_1.query(cdr).where(car).select();
};
exports.$$where = exports.$where(null, null);
//# sourceMappingURL=sequence.fn.js.map