// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import { parse } from './parser';
import { evaluate } from './evaluate';
import installCore from './operators/core';
import installArithmetic from './operators/arithmetic';
import installSequence from './operators/sequence';
import installJsx from './operators/jsx';
import installConcurrent from './operators/concurrent';
export const defaultReservedNames = {
    eval: '$eval',
    quote: '$quote',
    backquote: '$backquote',
    unquote: '$unquote',
    spread: '$spread',
    splice: '$splice',
    car: '$car',
    cdr: '$cdr',
    cons: '$cons',
    atom: '$atom',
    eq: '$eq',
    list: '$list',
    let: '$clisp-let',
    lambda: '$lambda',
    self: '$self',
    defun: '$defun',
    thiz: '$this',
    if: '$if',
    cond: '$cond',
    while: '$while',
    doWhile: '$do-while',
    until: '$until',
    doUntil: '$do-until',
    get: '$get',
    defvar: '$clisp-defvar',
    setq: '$clisp-setq',
    set: '$set',
    call: '$call',
    not: '$not',
    and: '$and',
    or: '$or',
    isSymbol: '$is-symbol',
    gensym: '$gensym',
    raise: '$raise',
    catch: '$catch',
    Template: 'Template',
};
export const defaultConfig = {
    raiseOnUnresolvedSymbol: false,
    enableEvaluate: true,
    enableHereDoc: true,
    enableSpread: true,
    enableSplice: true,
    enableShorthands: true,
    enableVerbatimStringLiteral: true,
    enableTailCallOptimization: true,
    enableRegExpMatchOperators: true,
    enableCompilationOperators: true,
    stripComments: false,
    wrapExternalValue: true,
    returnMultipleRoot: false,
    maxEvalCount: 0,
    reservedNames: defaultReservedNames,
    symbols: [],
    macros: [],
    funcs: [],
};
function initState(config, globals, strings, values) {
    return {
        strings: typeof strings === 'string' ? [strings] : strings,
        values: values || [],
        index: 0,
        pos: 0,
        line: 0,
        evalCount: 0,
        scopes: [{ isBlockLocal: false, scope: globals }],
        macroMap: new Map(config.macros.map(x => [x.name, x])),
        funcMap: new Map(config.funcs.map(x => [x.name, x])),
        symbolMap: new Map(config.symbols.map(x => [x.name, x])),
        config,
    };
}
function resetState(state, strings, values) {
    state.strings = typeof strings === 'string' ? [strings] : strings;
    state.values = values || [];
    state.index = 0;
    state.pos = 0;
    state.line = 0;
    state.evalCount = 0;
    return state;
}
export function SExpression(conf) {
    let config = conf || Object.assign({}, defaultConfig);
    let globalScope = {};
    let startup = [];
    const exec = (state, s) => {
        if (config.enableEvaluate) {
            for (let i = 0; i < s.length; i++) {
                s[i] = evaluate(state, s[i]);
            }
        }
        if (config.returnMultipleRoot) {
            return s.length === 1 ? s[0] : s;
        }
        else {
            return s[s.length - 1];
        }
    };
    const f = ((strings, ...values) => {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        return exec(state, startup.concat(parse(state)));
    });
    f.evaluateAST = (ast) => {
        const state = initState(config, Object.assign({}, globalScope), '');
        return exec(state, startup.concat(ast));
    };
    f.repl = () => {
        const state = initState(config, Object.assign({}, globalScope), '');
        exec(state, startup.slice(0));
        const fRepl = ((strings, ...values) => {
            resetState(state, strings, values);
            return exec(state, parse(state));
        });
        fRepl.sync = fRepl;
        return fRepl;
    };
    f.setGlobals = (globals) => {
        globalScope = Object.assign({}, globals || {});
        return f;
    };
    f.appendGlobals = (globals) => {
        globalScope = Object.assign({}, globalScope, globals || {});
        return f;
    };
    f.setStartup = (strings, ...values) => {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        startup = parse(state);
        return f;
    };
    f.setStartupAST = (ast) => {
        startup = ast;
        return f;
    };
    f.appendStartup = (strings, ...values) => {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        startup = startup.concat(parse(state));
        return f;
    };
    f.appendStartupAST = (ast) => {
        startup = startup.concat(ast);
        return f;
    };
    f.install = (installer) => {
        config = installer(config);
        return f;
    };
    return f;
}
export function SExpressionAsync(conf) {
    let config = conf || Object.assign({}, defaultConfig);
    let globalScope = {};
    let startup = [];
    const exec = (state, s) => __awaiter(this, void 0, void 0, function* () {
        if (config.enableEvaluate) {
            for (let i = 0; i < s.length; i++) {
                s[i] = evaluate(state, s[i]);
                if (typeof s[i] === 'object' && s[i] !== null && typeof s[i].then === 'function') {
                    s[i] = yield s[i];
                }
            }
        }
        if (config.returnMultipleRoot) {
            return s.length === 1 ? s[0] : s;
        }
        else {
            return s[s.length - 1];
        }
    });
    const f = ((strings, ...values) => __awaiter(this, void 0, void 0, function* () {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        return exec(state, startup.concat(parse(state)));
    }));
    f.evaluateAST = (ast) => {
        const state = initState(config, Object.assign({}, globalScope), '');
        return exec(state, startup.concat(ast));
    };
    f.repl = () => {
        const execSync = (stat, s) => {
            if (config.enableEvaluate) {
                for (let i = 0; i < s.length; i++) {
                    s[i] = evaluate(stat, s[i]);
                }
            }
            if (config.returnMultipleRoot) {
                return s.length === 1 ? s[0] : s;
            }
            else {
                return s[s.length - 1];
            }
        };
        const state = initState(config, Object.assign({}, globalScope), '');
        exec(state, startup.slice(0));
        const fRepl = ((strings, ...values) => __awaiter(this, void 0, void 0, function* () {
            resetState(state, strings, values);
            return exec(state, parse(state));
        }));
        const fReplSync = ((strings, ...values) => {
            resetState(state, strings, values);
            return execSync(state, parse(state));
        });
        fRepl.sync = fReplSync;
        return fRepl;
    };
    f.setGlobals = (globals) => {
        globalScope = Object.assign({}, globals || {});
        return f;
    };
    f.appendGlobals = (globals) => {
        globalScope = Object.assign({}, globalScope, globals || {});
        return f;
    };
    f.setStartup = (strings, ...values) => {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        startup = parse(state);
        return f;
    };
    f.setStartupAST = (ast) => {
        startup = ast;
        return f;
    };
    f.appendStartup = (strings, ...values) => {
        const state = initState(config, Object.assign({}, globalScope), strings, values);
        startup = startup.concat(parse(state));
        return f;
    };
    f.appendStartupAST = (ast) => {
        startup = startup.concat(ast);
        return f;
    };
    f.install = (installer) => {
        config = installer(config);
        return f;
    };
    return f;
}
export const S = (() => {
    const config = Object.assign({}, defaultConfig);
    config.enableEvaluate = false;
    config.returnMultipleRoot = true;
    return SExpression(config);
})();
export const L = (() => {
    let config = Object.assign({}, defaultConfig);
    config.reservedNames = Object.assign({}, config.reservedNames, {
        Template: '$concat',
    });
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config.stripComments = true;
    return SExpression(config);
})();
export const LS = L;
export const lisp = L;
// tslint:disable-next-line:variable-name
export const L_async = (() => {
    let config = Object.assign({}, defaultConfig);
    config.reservedNames = Object.assign({}, config.reservedNames, {
        Template: '$concat',
    });
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config.stripComments = true;
    return SExpressionAsync(config);
})();
// tslint:disable-next-line:variable-name
export const LS_async = L_async;
// tslint:disable-next-line:variable-name
export const lisp_async = L_async;
export const LM = (() => {
    let config = Object.assign({}, defaultConfig);
    config.reservedNames = Object.assign({}, config.reservedNames, {
        Template: '$concat',
    });
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config.stripComments = true;
    config.returnMultipleRoot = true;
    return SExpression(config);
})();
// tslint:disable-next-line:variable-name
export const LM_async = (() => {
    let config = Object.assign({}, defaultConfig);
    config.reservedNames = Object.assign({}, config.reservedNames, {
        Template: '$concat',
    });
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config.stripComments = true;
    config.returnMultipleRoot = true;
    return SExpressionAsync(config);
})();
export function LSX(lsxConf) {
    let config = Object.assign({}, defaultConfig);
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config = installJsx(config, lsxConf);
    config.stripComments = true;
    return SExpression(config);
}
export function LSX_async(lsxConf) {
    let config = Object.assign({}, defaultConfig);
    config = installCore(config);
    config = installArithmetic(config);
    config = installSequence(config);
    config = installConcurrent(config);
    config = installJsx(config, lsxConf);
    config.stripComments = true;
    return SExpressionAsync(config);
}
//# sourceMappingURL=s-expression.js.map